@@ -17,18 +17,16 @@
  */
 
 #include "config.h"
+#include "ass_compat.h"
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <strings.h>
 #include <assert.h>
 #include <errno.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <unistd.h>
 #include <inttypes.h>
-#include <ctype.h>
 
 #ifdef CONFIG_ICONV
 #include <iconv.h>
@@ -37,6 +35,7 @@
 #include "ass.h"
 #include "ass_utils.h"
 #include "ass_library.h"
+#include "ass_string.h"
 
 #define ass_atof(STR) (ass_strtod((STR),NULL))
 
@@ -183,12 +182,12 @@ static long long string2timecode(ASS_Library *library, char *p)
 {
     unsigned h, m, s, ms;
     long long tm;
-    int res = sscanf(p, "%1d:%2d:%2d.%2d", &h, &m, &s, &ms);
+    int res = sscanf(p, "%d:%d:%d.%d", &h, &m, &s, &ms);
     if (res < 4) {
         ass_msg(library, MSGL_WARN, "Bad timestamp");
         return 0;
     }
-    tm = ((h * 60 + m) * 60 + s) * 1000 + ms * 10;
+    tm = ((h * 60LL + m) * 60 + s) * 1000 + ms * 10;
     return tm;
 }
 
@@ -212,7 +211,7 @@ static int numpad2align(int val)
 
 
 #define ALIAS(alias,name) \
-        if (strcasecmp(tname, #alias) == 0) {tname = #name;}
+        if (ass_strcasecmp(tname, #alias) == 0) {tname = #name;}
 
 /* One section started with PARSE_START and PARSE_END parses a single token
  * (contained in the variable named token) for the header indicated by the
@@ -228,32 +227,29 @@ static int numpad2align(int val)
 #define PARSE_END   }
 
 #define ANYVAL(name,func) \
-	} else if (strcasecmp(tname, #name) == 0) { \
+	} else if (ass_strcasecmp(tname, #name) == 0) { \
 		target->name = func(token);
 
 #define STRVAL(name) \
-	} else if (strcasecmp(tname, #name) == 0) { \
+	} else if (ass_strcasecmp(tname, #name) == 0) { \
 		if (target->name != NULL) free(target->name); \
 		target->name = strdup(token);
 
 #define STARREDSTRVAL(name) \
-    } else if (strcasecmp(tname, #name) == 0) { \
+    } else if (ass_strcasecmp(tname, #name) == 0) { \
         if (target->name != NULL) free(target->name); \
         while (*token == '*') ++token; \
         target->name = strdup(token);
 
-#define COLORVAL(name) \
-	} else if (strcasecmp(tname, #name) == 0) { \
-		target->name = string2color(track->library, token, 0);
-
+#define COLORVAL(name) ANYVAL(name,parse_color_header)
 #define INTVAL(name) ANYVAL(name,atoi)
 #define FPVAL(name) ANYVAL(name,ass_atof)
 #define TIMEVAL(name) \
-	} else if (strcasecmp(tname, #name) == 0) { \
+	} else if (ass_strcasecmp(tname, #name) == 0) { \
 		target->name = string2timecode(track->library, token);
 
 #define STYLEVAL(name) \
-	} else if (strcasecmp(tname, #name) == 0) { \
+	} else if (ass_strcasecmp(tname, #name) == 0) { \
 		target->name = lookup_style(track, token);
 
 static char *next_token(char **str)
@@ -312,7 +308,7 @@ static int process_event_tail(ASS_Track *track, ASS_Event *event,
 
     while (1) {
         NEXT(q, tname);
-        if (strcasecmp(tname, "Text") == 0) {
+        if (ass_strcasecmp(tname, "Text") == 0) {
             char *last;
             event->Text = strdup(p);
             if (*event->Text != 0) {
@@ -365,19 +361,19 @@ void ass_process_force_style(ASS_Track *track)
         *eq = '\0';
         token = eq + 1;
 
-        if (!strcasecmp(*fs, "PlayResX"))
+        if (!ass_strcasecmp(*fs, "PlayResX"))
             track->PlayResX = atoi(token);
-        else if (!strcasecmp(*fs, "PlayResY"))
+        else if (!ass_strcasecmp(*fs, "PlayResY"))
             track->PlayResY = atoi(token);
-        else if (!strcasecmp(*fs, "Timer"))
+        else if (!ass_strcasecmp(*fs, "Timer"))
             track->Timer = ass_atof(token);
-        else if (!strcasecmp(*fs, "WrapStyle"))
+        else if (!ass_strcasecmp(*fs, "WrapStyle"))
             track->WrapStyle = atoi(token);
-        else if (!strcasecmp(*fs, "ScaledBorderAndShadow"))
+        else if (!ass_strcasecmp(*fs, "ScaledBorderAndShadow"))
             track->ScaledBorderAndShadow = parse_bool(token);
-        else if (!strcasecmp(*fs, "Kerning"))
+        else if (!ass_strcasecmp(*fs, "Kerning"))
             track->Kerning = parse_bool(token);
-        else if (!strcasecmp(*fs, "YCbCr Matrix"))
+        else if (!ass_strcasecmp(*fs, "YCbCr Matrix"))
             track->YCbCrMatrix = parse_ycbcr_matrix(token);
 
         dt = strrchr(*fs, '.');
@@ -391,7 +387,7 @@ void ass_process_force_style(ASS_Track *track)
         }
         for (sid = 0; sid < track->n_styles; ++sid) {
             if (style == NULL
-                || strcasecmp(track->styles[sid].Name, style) == 0) {
+                || ass_strcasecmp(track->styles[sid].Name, style) == 0) {
                 target = track->styles + sid;
                 PARSE_START
                     STRVAL(FontName)
@@ -578,10 +574,8 @@ static int process_info_line(ASS_Track *track, char *str)
         track->YCbCrMatrix = parse_ycbcr_matrix(str + 13);
     } else if (!strncmp(str, "Language:", 9)) {
         char *p = str + 9;
-        while (*p && isspace(*p)) p++;
-        track->Language = malloc(3);
-        strncpy(track->Language, p, 2);
-        track->Language[2] = 0;
+        while (*p && ass_isspace(*p)) p++;
+        track->Language = strndup(p, 2);
     }
     return 0;
 }
@@ -669,6 +663,8 @@ static int decode_font(ASS_Track *track)
         goto error_decode_font;
     }
     buf = malloc(size / 4 * 3 + 2);
+    if (!buf)
+        goto error_decode_font;
     q = buf;
     for (i = 0, p = (unsigned char *) track->parser_priv->fontdata;
          i < size / 4; i++, p += 4) {
@@ -746,17 +742,17 @@ static int process_fonts_line(ASS_Track *track, char *str)
 */
 static int process_line(ASS_Track *track, char *str)
 {
-    if (!strncasecmp(str, "[Script Info]", 13)) {
+    if (!ass_strncasecmp(str, "[Script Info]", 13)) {
         track->parser_priv->state = PST_INFO;
-    } else if (!strncasecmp(str, "[V4 Styles]", 11)) {
+    } else if (!ass_strncasecmp(str, "[V4 Styles]", 11)) {
         track->parser_priv->state = PST_STYLES;
         track->track_type = TRACK_TYPE_SSA;
-    } else if (!strncasecmp(str, "[V4+ Styles]", 12)) {
+    } else if (!ass_strncasecmp(str, "[V4+ Styles]", 12)) {
         track->parser_priv->state = PST_STYLES;
         track->track_type = TRACK_TYPE_ASS;
-    } else if (!strncasecmp(str, "[Events]", 8)) {
+    } else if (!ass_strncasecmp(str, "[Events]", 8)) {
         track->parser_priv->state = PST_EVENTS;
-    } else if (!strncasecmp(str, "[Fonts]", 7)) {
+    } else if (!ass_strncasecmp(str, "[Fonts]", 7)) {
         track->parser_priv->state = PST_FONTS;
     } else {
         switch (track->parser_priv->state) {
@@ -776,12 +772,6 @@ static int process_line(ASS_Track *track, char *str)
             break;
         }
     }
-
-    // there is no explicit end-of-font marker in ssa/ass
-    if ((track->parser_priv->state != PST_FONTS)
-        && (track->parser_priv->fontname))
-        decode_font(track);
-
     return 0;
 }
 
@@ -809,6 +799,9 @@ static int process_text(ASS_Track *track, char *str)
             break;
         p = q;
     }
+    // there is no explicit end-of-font marker in ssa/ass
+    if (track->parser_priv->fontname)
+        decode_font(track);
     return 0;
 }
 
@@ -821,6 +814,8 @@ static int process_text(ASS_Track *track, char *str)
 void ass_process_data(ASS_Track *track, char *data, int size)
 {
     char *str = malloc(size + 1);
+    if (!str)
+        return;
 
     memcpy(str, data, size);
     str[size] = '\0';
@@ -881,6 +876,8 @@ void ass_process_chunk(ASS_Track *track, char *data, int size,
     }
 
     str = malloc(size + 1);
+    if (!str)
+        return;
     memcpy(str, data, size);
     str[size] = '\0';
     ass_msg(track->library, MSGL_V, "Event at %" PRId64 ", +%" PRId64 ": %s",
@@ -944,27 +941,11 @@ static char *sub_recode(ASS_Library *library, char *data, size_t size,
     char *outbuf;
     assert(codepage);
 
-    {
-        const char *cp_tmp = codepage;
-#ifdef CONFIG_ENCA
-        char enca_lang[3], enca_fallback[100];
-        if (sscanf(codepage, "enca:%2s:%99s", enca_lang, enca_fallback) == 2
-            || sscanf(codepage, "ENCA:%2s:%99s", enca_lang,
-                      enca_fallback) == 2) {
-            cp_tmp =
-                ass_guess_buffer_cp(library, (unsigned char *) data, size,
-                                    enca_lang, enca_fallback);
-        }
-#endif
-        if ((icdsc = iconv_open(tocp, cp_tmp)) != (iconv_t) (-1)) {
-            ass_msg(library, MSGL_V, "Opened iconv descriptor");
-        } else
-            ass_msg(library, MSGL_ERR, "Error opening iconv descriptor");
-#ifdef CONFIG_ENCA
-        if (cp_tmp != codepage) {
-            free((void*)cp_tmp);
-        }
-#endif
+    if ((icdsc = iconv_open(tocp, codepage)) != (iconv_t) (-1)) {
+        ass_msg(library, MSGL_V, "Opened iconv descriptor");
+    } else {
+        ass_msg(library, MSGL_ERR, "Error opening iconv descriptor");
+        return NULL;
     }
 
     {
@@ -977,6 +958,8 @@ static char *sub_recode(ASS_Library *library, char *data, size_t size,
         int clear = 0;
 
         outbuf = malloc(osize);
+        if (!outbuf)
+            goto out;
         ip = data;
         op = outbuf;
 
@@ -990,7 +973,13 @@ static char *sub_recode(ASS_Library *library, char *data, size_t size,
             if (rc == (size_t) (-1)) {
                 if (errno == E2BIG) {
                     size_t offset = op - outbuf;
-                    outbuf = (char *) realloc(outbuf, osize + size);
+                    char *nbuf = realloc(outbuf, osize + size);
+                    if (!nbuf) {
+                        free(outbuf);
+                        outbuf = 0;
+                        goto out;
+                    }
+                    outbuf = nbuf;
                     op = outbuf + offset;
                     osize += size;
                     oleft += size;
@@ -1009,7 +998,6 @@ static char *sub_recode(ASS_Library *library, char *data, size_t size,
 out:
     if (icdsc != (iconv_t) (-1)) {
         (void) iconv_close(icdsc);
-        icdsc = (iconv_t) (-1);
         ass_msg(library, MSGL_V, "Closed iconv descriptor");
     }
 
@@ -1023,7 +1011,7 @@ out:
  * \param bufsize out: file size
  * \return pointer to file contents. Caller is responsible for its deallocation.
  */
-static char *read_file(ASS_Library *library, char *fname, size_t *bufsize)
+char *read_file(ASS_Library *library, char *fname, size_t *bufsize)
 {
     int res;
     long sz;
@@ -1049,7 +1037,11 @@ static char *read_file(ASS_Library *library, char *fname, size_t *bufsize)
 
     ass_msg(library, MSGL_V, "File size: %ld", sz);
 
-    buf = malloc(sz + 1);
+    buf = sz < SIZE_MAX ? malloc(sz + 1) : NULL;
+    if (!buf) {
+        fclose(fp);
+        return NULL;
+    }
     assert(buf);
     bytes_read = 0;
     do {
@@ -1088,10 +1080,6 @@ static ASS_Track *parse_memory(ASS_Library *library, char *buf)
     for (i = 0; i < track->n_events; ++i)
         track->events[i].ReadOrder = i;
 
-    // there is no explicit end-of-font marker in ssa/ass
-    if (track->parser_priv->fontname)
-        decode_font(track);
-
     if (track->track_type == TRACK_TYPE_UNKNOWN) {
         ass_free_track(track);
         return 0;
@@ -1220,12 +1208,13 @@ int ass_read_styles(ASS_Track *track, char *fname, char *codepage)
         buf = tmpbuf;
     }
     if (!buf)
-        return 0;
+        return 1;
 #endif
 
     old_state = track->parser_priv->state;
     track->parser_priv->state = PST_STYLES;
     process_text(track, buf);
+    free(buf);
     track->parser_priv->state = old_state;
 
     return 0;
@@ -1278,9 +1267,15 @@ long long ass_step_sub(ASS_Track *track, long long now, int movement)
 ASS_Track *ass_new_track(ASS_Library *library)
 {
     ASS_Track *track = calloc(1, sizeof(ASS_Track));
+    if (!track)
+        return NULL;
     track->library = library;
     track->ScaledBorderAndShadow = 1;
     track->parser_priv = calloc(1, sizeof(ASS_ParserPriv));
+    if (!track->parser_priv) {
+        free(track);
+        return NULL;
+    }
     return track;
 }
 
