@@ -17,6 +17,7 @@
  */
 
 #include "config.h"
+#include "ass_compat.h"
 
 #include <inttypes.h>
 #include <ft2build.h>
@@ -25,13 +26,12 @@
 #include FT_GLYPH_H
 #include FT_TRUETYPE_TABLES_H
 #include FT_OUTLINE_H
-#include <strings.h>
 #include <limits.h>
 
 #include "ass.h"
 #include "ass_library.h"
 #include "ass_font.h"
-#include "ass_fontconfig.h"
+#include "ass_fontselect.h"
 #include "ass_utils.h"
 #include "ass_shaper.h"
 
@@ -39,7 +39,7 @@
  * Select a good charmap, prefer Microsoft Unicode charmaps.
  * Otherwise, let FreeType decide.
  */
-static void charmap_magic(ASS_Library *library, FT_Face face)
+void charmap_magic(ASS_Library *library, FT_Face face)
 {
     int i;
     int ms_cmap = -1;
@@ -92,18 +92,6 @@ uint32_t ass_font_index_magic(FT_Face face, uint32_t symbol)
     }
 }
 
-/**
- * \brief find a memory font by name
- */
-static int find_font(ASS_Library *library, char *name)
-{
-    int i;
-    for (i = 0; i < library->num_fontdata; ++i)
-        if (strcasecmp(name, library->fontdata[i].name) == 0)
-            return i;
-    return -1;
-}
-
 static void buggy_font_workaround(FT_Face face)
 {
     // Some fonts have zero Ascender/Descender fields in 'hhea' table.
@@ -123,57 +111,108 @@ static void buggy_font_workaround(FT_Face face)
     }
 }
 
+static unsigned long
+read_stream_font(FT_Stream stream, unsigned long offset, unsigned char *buffer,
+                 unsigned long count)
+{
+    ASS_FontStream *font = (ASS_FontStream *)stream->descriptor.pointer;
+
+    font->func(font->priv, buffer, offset, count);
+    return count;
+}
+
+static void
+close_stream_font(FT_Stream stream)
+{
+    free(stream->descriptor.pointer);
+    free(stream);
+}
+
 /**
  * \brief Select a face with the given charcode and add it to ASS_Font
  * \return index of the new face in font->faces, -1 if failed
  */
-static int add_face(void *fc_priv, ASS_Font *font, uint32_t ch)
+static int add_face(ASS_FontSelector *fontsel, ASS_Font *font, uint32_t ch)
 {
     char *path;
-    int index;
+    char *postscript_name = NULL;
+    int i, index, uid, error;
+    ASS_FontStream stream = { NULL, NULL };
     FT_Face face;
-    int error;
-    int mem_idx;
 
     if (font->n_faces == ASS_FONT_MAX_FACES)
         return -1;
 
-    path =
-        fontconfig_select(font->library, fc_priv, font->desc.family,
-                          font->desc.treat_family_as_pattern,
-                          font->desc.bold, font->desc.italic, &index, ch);
+    path = ass_font_select(fontsel, font->library, font , &index,
+            &postscript_name, &uid, &stream, ch);
+
     if (!path)
         return -1;
 
-    mem_idx = find_font(font->library, path);
-    if (mem_idx >= 0) {
-        error =
-            FT_New_Memory_Face(font->ftlibrary,
-                               (unsigned char *) font->library->
-                               fontdata[mem_idx].data,
-                               font->library->fontdata[mem_idx].size, index,
-                               &face);
+    for (i = 0; i < font->n_faces; i++) {
+        if (font->faces_uid[i] == uid) {
+            ass_msg(font->library, MSGL_INFO,
+                    "Got a font face that already is available! Skipping.");
+            return i;
+        }
+    }
+
+    if (stream.func) {
+        FT_Open_Args args;
+        FT_Stream ftstream = calloc(1, sizeof(FT_StreamRec));
+        ASS_FontStream *fs  = calloc(1, sizeof(ASS_FontStream));
+
+        *fs = stream;
+        ftstream->size  = stream.func(stream.priv, NULL, 0, 0);
+        ftstream->read  = read_stream_font;
+        ftstream->close = close_stream_font;
+        ftstream->descriptor.pointer = (void *)fs;
+
+        memset(&args, 0, sizeof(FT_Open_Args));
+        args.flags  = FT_OPEN_STREAM;
+        args.stream = ftstream;
+
+        error = FT_Open_Face(font->ftlibrary, &args, index, &face);
+
         if (error) {
             ass_msg(font->library, MSGL_WARN,
                     "Error opening memory font: '%s'", path);
-            free(path);
             return -1;
         }
+
     } else {
         error = FT_New_Face(font->ftlibrary, path, index, &face);
         if (error) {
             ass_msg(font->library, MSGL_WARN,
                     "Error opening font: '%s', %d", path, index);
-            free(path);
             return -1;
         }
+
+        if (postscript_name && index < 0 && face->num_faces > 0) {
+            // The font provider gave us a post_script name and is not sure
+            // about the face index.. so use the postscript name to find the
+            // correct face_index in the collection!
+            for (int i = 0; i < face->num_faces; i++) {
+                FT_Done_Face(face);
+                error = FT_New_Face(font->ftlibrary, path, i, &face);
+                if (error) {
+                    ass_msg(font->library, MSGL_WARN,
+                            "Error opening font: '%s', %d", path, i);
+                    return -1;
+                }
+
+                if (strcmp(FT_Get_Postscript_Name(face), postscript_name) == 0)
+                    break;
+            }
+        }
     }
+
     charmap_magic(font->library, face);
     buggy_font_workaround(face);
 
-    font->faces[font->n_faces++] = face;
+    font->faces[font->n_faces] = face;
+    font->faces_uid[font->n_faces++] = uid;
     ass_face_set_size(face, font->size);
-    free(path);
     return font->n_faces - 1;
 }
 
@@ -181,7 +220,7 @@ static int add_face(void *fc_priv, ASS_Font *font, uint32_t ch)
  * \brief Create a new ASS_Font according to "desc" argument
  */
 ASS_Font *ass_font_new(Cache *font_cache, ASS_Library *library,
-                       FT_Library ftlibrary, void *fc_priv,
+                       FT_Library ftlibrary, ASS_FontSelector *fontsel,
                        ASS_FontDesc *desc)
 {
     int error;
@@ -197,7 +236,6 @@ ASS_Font *ass_font_new(Cache *font_cache, ASS_Library *library,
     font.shaper_priv = NULL;
     font.n_faces = 0;
     font.desc.family = strdup(desc->family);
-    font.desc.treat_family_as_pattern = desc->treat_family_as_pattern;
     font.desc.bold = desc->bold;
     font.desc.italic = desc->italic;
     font.desc.vertical = desc->vertical;
@@ -206,7 +244,7 @@ ASS_Font *ass_font_new(Cache *font_cache, ASS_Library *library,
     font.v.x = font.v.y = 0;
     font.size = 0.;
 
-    error = add_face(fc_priv, &font, 0);
+    error = add_face(fontsel, &font, 0);
     if (error == -1) {
         free(font.desc.family);
         return 0;
@@ -338,25 +376,26 @@ static int ass_strike_outline_glyph(FT_Face face, ASS_Font *font,
     TT_OS2 *os2 = FT_Get_Sfnt_Table(face, ft_sfnt_os2);
     TT_Postscript *ps = FT_Get_Sfnt_Table(face, ft_sfnt_post);
     FT_Outline *ol = &((FT_OutlineGlyph) glyph)->outline;
-    int bear, advance, y_scale, i, dir;
+    int advance, y_scale, i, dir;
 
     if (!under && !through)
         return 0;
 
     // Grow outline
     i = (under ? 4 : 0) + (through ? 4 : 0);
-    ol->points = realloc(ol->points, sizeof(FT_Vector) *
-                         (ol->n_points + i));
-    ol->tags = realloc(ol->tags, ol->n_points + i);
+    if (ol->n_points > SHRT_MAX - i)
+        return 0;
+    if (!ASS_REALLOC_ARRAY(ol->points, ol->n_points + i))
+        return 0;
+    if (!ASS_REALLOC_ARRAY(ol->tags, ol->n_points + i))
+        return 0;
     i = !!under + !!through;
-    ol->contours = realloc(ol->contours, sizeof(short) *
-                           (ol->n_contours + i));
+    if (ol->n_contours > SHRT_MAX - i)
+        return 0;
+    if (!ASS_REALLOC_ARRAY(ol->contours, ol->n_contours + i))
+        return 0;
 
-    // If the bearing is negative, the glyph starts left of the current
-    // pen position
-    bear = FFMIN(face->glyph->metrics.horiBearingX, 0);
-    // We're adding half a pixel to avoid small gaps
-    advance = d16_to_d6(glyph->advance.x) + 32;
+    advance = d16_to_d6(glyph->advance.x);
     y_scale = face->size->metrics.y_scale;
 
     // Reverse drawing direction for non-truetype fonts
@@ -364,46 +403,97 @@ static int ass_strike_outline_glyph(FT_Face face, ASS_Font *font,
 
     // Add points to the outline
     if (under && ps) {
-        int pos = FT_MulFix(ps->underlinePosition, y_scale * font->scale_y);
-        int size = FT_MulFix(ps->underlineThickness,
-                             y_scale * font->scale_y / 2);
+        int pos = FT_MulFix(ps->underlinePosition, y_scale);
+        int size = FT_MulFix(ps->underlineThickness, y_scale / 2);
 
         if (pos > 0 || size <= 0)
             return 1;
 
-        add_line(ol, bear, advance, dir, pos, size);
+        add_line(ol, 0, advance, dir, pos, size);
     }
 
     if (through && os2) {
-        int pos = FT_MulFix(os2->yStrikeoutPosition, y_scale * font->scale_y);
-        int size = FT_MulFix(os2->yStrikeoutSize, y_scale * font->scale_y / 2);
+        int pos = FT_MulFix(os2->yStrikeoutPosition, y_scale);
+        int size = FT_MulFix(os2->yStrikeoutSize, y_scale / 2);
 
         if (pos < 0 || size <= 0)
             return 1;
 
-        add_line(ol, bear, advance, dir, pos, size);
+        add_line(ol, 0, advance, dir, pos, size);
     }
 
     return 0;
 }
 
-void outline_copy(FT_Library lib, FT_Outline *source, FT_Outline **dest)
-{
-    if (source == NULL) {
-        *dest = NULL;
-        return;
-    }
-    *dest = calloc(1, sizeof(**dest));
 
-    FT_Outline_New(lib, source->n_points, source->n_contours, *dest);
-    FT_Outline_Copy(source, *dest);
+int outline_alloc(ASS_Outline *outline, size_t n_points, size_t n_contours)
+{
+    outline->contours = malloc(sizeof(size_t) * n_contours);
+    outline->points = malloc(sizeof(FT_Vector) * n_points);
+    outline->tags = malloc(n_points);
+    if (!outline->contours || !outline->points || !outline->tags)
+        return 0;
+
+    outline->max_contours = n_contours;
+    outline->max_points = n_points;
+    return 1;
 }
 
-void outline_free(FT_Library lib, FT_Outline *outline)
+ASS_Outline *outline_convert(const FT_Outline *source)
 {
-    if (outline)
-        FT_Outline_Done(lib, outline);
-    free(outline);
+    if (!source)
+        return NULL;
+
+    ASS_Outline *ol = calloc(1, sizeof(*ol));
+    if (!ol)
+        return NULL;
+
+    if (!outline_alloc(ol, source->n_points, source->n_contours)) {
+        outline_free(ol);
+        free(ol);
+        return NULL;
+    }
+
+    for (int i = 0; i < source->n_contours; ++i)
+        ol->contours[i] = source->contours[i];
+    memcpy(ol->points, source->points, sizeof(FT_Vector) * source->n_points);
+    memcpy(ol->tags, source->tags, source->n_points);
+    ol->n_contours = source->n_contours;
+    ol->n_points = source->n_points;
+    return ol;
+}
+
+ASS_Outline *outline_copy(const ASS_Outline *source)
+{
+    if (!source)
+        return NULL;
+
+    ASS_Outline *ol = calloc(1, sizeof(*ol));
+    if (!ol)
+        return NULL;
+
+    if (!outline_alloc(ol, source->n_points, source->n_contours)) {
+        outline_free(ol);
+        free(ol);
+        return NULL;
+    }
+
+    memcpy(ol->contours, source->contours, sizeof(size_t) * source->n_contours);
+    memcpy(ol->points, source->points, sizeof(FT_Vector) * source->n_points);
+    memcpy(ol->tags, source->tags, source->n_points);
+    ol->n_contours = source->n_contours;
+    ol->n_points = source->n_points;
+    return ol;
+}
+
+void outline_free(ASS_Outline *outline)
+{
+    if (!outline)
+        return;
+
+    free(outline->contours);
+    free(outline->points);
+    free(outline->tags);
 }
 
 /**
@@ -427,8 +517,8 @@ static void ass_glyph_embolden(FT_GlyphSlot slot)
  * Finds a face that has the requested codepoint and returns both face
  * and glyph index.
  */
-int ass_font_get_index(void *fcpriv, ASS_Font *font, uint32_t symbol,
-                       int *face_index, int *glyph_index)
+int ass_font_get_index(ASS_FontSelector *fontsel, ASS_Font *font,
+                       uint32_t symbol, int *face_index, int *glyph_index)
 {
     int index = 0;
     int i;
@@ -462,14 +552,13 @@ int ass_font_get_index(void *fcpriv, ASS_Font *font, uint32_t symbol,
             *face_index = i;
     }
 
-#ifdef CONFIG_FONTCONFIG
     if (index == 0) {
         int face_idx;
         ass_msg(font->library, MSGL_INFO,
                 "Glyph 0x%X not found, selecting one more "
                 "font for (%s, %d, %d)", symbol, font->desc.family,
                 font->desc.bold, font->desc.italic);
-        face_idx = *face_index = add_face(fcpriv, font, symbol);
+        face_idx = *face_index = add_face(fontsel, font, symbol);
         if (face_idx >= 0) {
             face = font->faces[face_idx];
             index = FT_Get_Char_Index(face, ass_font_index_magic(face, symbol));
@@ -490,7 +579,7 @@ int ass_font_get_index(void *fcpriv, ASS_Font *font, uint32_t symbol,
             }
         }
     }
-#endif
+
     // FIXME: make sure we have a valid face_index. this is a HACK.
     *face_index  = FFMAX(*face_index, 0);
     *glyph_index = index;
@@ -502,9 +591,8 @@ int ass_font_get_index(void *fcpriv, ASS_Font *font, uint32_t symbol,
  * \brief Get a glyph
  * \param ch character code
  **/
-FT_Glyph ass_font_get_glyph(void *fontconfig_priv, ASS_Font *font,
-                            uint32_t ch, int face_index, int index,
-                            ASS_Hinting hinting, int deco)
+FT_Glyph ass_font_get_glyph(ASS_Font *font, uint32_t ch, int face_index,
+                            int index, ASS_Hinting hinting, int deco)
 {
     int error;
     FT_Glyph glyph;
@@ -540,7 +628,7 @@ FT_Glyph ass_font_get_glyph(void *fontconfig_priv, ASS_Font *font,
     }
 
     if (!(face->style_flags & FT_STYLE_FLAG_BOLD) &&
-        (font->desc.bold > 80)) {
+        (font->desc.bold > 400)) {
         ass_glyph_embolden(face->glyph);
     }
     error = FT_Get_Glyph(face->glyph, &glyph);
@@ -566,6 +654,9 @@ FT_Glyph ass_font_get_glyph(void *fontconfig_priv, ASS_Font *font,
         glyph->advance.x = face->glyph->linearVertAdvance;
     }
 
+    ass_strike_outline_glyph(face, font, glyph, deco & DECO_UNDERLINE,
+                             deco & DECO_STRIKETHROUGH);
+
     // Apply scaling and shift
     FT_Matrix scale = { double_to_d16(font->scale_x), 0, 0,
                         double_to_d16(font->scale_y) };
@@ -574,9 +665,6 @@ FT_Glyph ass_font_get_glyph(void *fontconfig_priv, ASS_Font *font,
     FT_Outline_Translate(outl, font->v.x, font->v.y);
     glyph->advance.x *= font->scale_x;
 
-    ass_strike_outline_glyph(face, font, glyph, deco & DECO_UNDERLINE,
-                             deco & DECO_STRIKETHROUGH);
-
     return glyph;
 }
 
@@ -612,11 +700,12 @@ FT_Vector ass_font_get_kerning(ASS_Font *font, uint32_t c1, uint32_t c2)
 void ass_font_free(ASS_Font *font)
 {
     int i;
-    for (i = 0; i < font->n_faces; ++i)
-        if (font->faces[i])
-            FT_Done_Face(font->faces[i]);
     if (font->shaper_priv)
         ass_shaper_font_data_free(font->shaper_priv);
+    for (i = 0; i < font->n_faces; ++i) {
+        if (font->faces[i])
+            FT_Done_Face(font->faces[i]);
+    }
     free(font->desc.family);
     free(font);
 }
@@ -640,22 +729,57 @@ get_contour_cbox(FT_BBox *box, FT_Vector *points, int start, int end)
 }
 
 /**
- * \brief Determine winding direction of a contour
- * \return direction; 0 = clockwise
+ * \brief Determine signed area of a contour
+ * \return area doubled
  */
-static int get_contour_direction(FT_Vector *points, int start, int end)
+static long long get_contour_area(FT_Vector *points, int start, int end)
 {
-    int i;
-    long long sum = 0;
-    int x = points[start].x;
-    int y = points[start].y;
-    for (i = start + 1; i <= end; i++) {
-        sum += x * (points[i].y - y) - y * (points[i].x - x);
+    long long area = 0;
+    int x = points[end].x;
+    int y = points[end].y;
+    for (int i = start; i <= end; i++) {
+        area += (long long)(points[i].x + x) * (points[i].y - y);
         x = points[i].x;
         y = points[i].y;
     }
-    sum += x * (points[start].y - y) - y * (points[start].x - x);
-    return sum > 0;
+    return area;
+}
+
+void outline_translate(const ASS_Outline *outline, FT_Pos dx, FT_Pos dy)
+{
+    for (size_t i = 0; i < outline->n_points; ++i) {
+        outline->points[i].x += dx;
+        outline->points[i].y += dy;
+    }
+}
+
+void outline_transform(const ASS_Outline *outline, const FT_Matrix *matrix)
+{
+    for (size_t i = 0; i < outline->n_points; ++i) {
+        FT_Pos x = FT_MulFix(outline->points[i].x, matrix->xx) +
+                   FT_MulFix(outline->points[i].y, matrix->xy);
+        FT_Pos y = FT_MulFix(outline->points[i].x, matrix->yx) +
+                   FT_MulFix(outline->points[i].y, matrix->yy);
+        outline->points[i].x = x;
+        outline->points[i].y = y;
+    }
+}
+
+void outline_get_cbox(const ASS_Outline *outline, FT_BBox *cbox)
+{
+    if (!outline->n_points) {
+        cbox->xMin = cbox->xMax = 0;
+        cbox->yMin = cbox->yMax = 0;
+        return;
+    }
+    cbox->xMin = cbox->xMax = outline->points[0].x;
+    cbox->yMin = cbox->yMax = outline->points[0].y;
+    for (size_t i = 1; i < outline->n_points; ++i) {
+        cbox->xMin = FFMIN(cbox->xMin, outline->points[i].x);
+        cbox->xMax = FFMAX(cbox->xMax, outline->points[i].x);
+        cbox->yMin = FFMIN(cbox->yMin, outline->points[i].y);
+        cbox->yMax = FFMAX(cbox->yMax, outline->points[i].y);
+    }
 }
 
 /**
@@ -674,7 +798,7 @@ static int get_contour_direction(FT_Vector *points, int start, int end)
  * \param border_x border size, x direction, d6 format
  * \param border_x border size, y direction, d6 format
  */
-void fix_freetype_stroker(FT_Outline *outline, int border_x, int border_y)
+void fix_freetype_stroker(ASS_Outline *outline, int border_x, int border_y)
 {
     int nc = outline->n_contours;
     int begin, stop;
@@ -684,17 +808,16 @@ void fix_freetype_stroker(FT_Outline *outline, int border_x, int border_y)
     int end = -1;
     FT_BBox *boxes = malloc(nc * sizeof(FT_BBox));
     int i, j;
-    int inside_direction;
-
-    inside_direction = FT_Outline_Get_Orientation(outline) ==
-        FT_ORIENTATION_TRUETYPE;
 
+    long long area = 0;
     // create a list of cboxes of the contours
     for (i = 0; i < nc; i++) {
         start = end + 1;
         end = outline->contours[i];
         get_contour_cbox(&boxes[i], outline->points, start, end);
+        area += get_contour_area(outline->points, start, end);
     }
+    int inside_direction = area < 0;
 
     // for each contour, check direction and whether it's "outside"
     // or contained in another contour
@@ -702,7 +825,7 @@ void fix_freetype_stroker(FT_Outline *outline, int border_x, int border_y)
     for (i = 0; i < nc; i++) {
         start = end + 1;
         end = outline->contours[i];
-        int dir = get_contour_direction(outline->points, start, end);
+        int dir = get_contour_area(outline->points, start, end) > 0;
         valid_cont[i] = 1;
         if (dir == inside_direction) {
             for (j = 0; j < nc; j++) {
