@@ -30,15 +30,12 @@
 #include "hb.h"
 #endif
 
-// XXX: fix the inclusion mess so we can avoid doing this here
-typedef struct ass_shaper ASS_Shaper;
-
 #include "ass.h"
 #include "ass_font.h"
 #include "ass_bitmap.h"
 #include "ass_cache.h"
 #include "ass_utils.h"
-#include "ass_fontconfig.h"
+#include "ass_fontselect.h"
 #include "ass_library.h"
 #include "ass_drawing.h"
 #include "ass_bitmap.h"
@@ -85,7 +82,7 @@ typedef struct {
     double par;                 // user defined pixel aspect ratio (0 = unset)
     ASS_Hinting hinting;
     ASS_ShapingLevel shaper;
-    int selective_style_overrides;
+    int selective_style_overrides; // ASS_OVERRIDE_* flags
 
     char *default_font;
     char *default_family;
@@ -107,48 +104,29 @@ typedef enum {
     EF_KARAOKE_KO
 } Effect;
 
+typedef struct
+{
+    int x_min, y_min, x_max, y_max;
+} Rectangle;
+
 // describes a combined bitmap
 typedef struct {
-    Bitmap *bm;                 // glyphs bitmap
-    unsigned w;
-    unsigned h;
-    Bitmap *bm_o;               // outline bitmap
-    unsigned o_w;
-    unsigned o_h;
-    Bitmap *bm_s;               // shadow bitmap
-    FT_Vector pos;
+    FilterDesc filter;
     uint32_t c[4];              // colors
-    FT_Vector advance;          // 26.6
     Effect effect_type;
     int effect_timing;          // time duration of current karaoke word
     // after process_karaoke_effects: distance in pixels from the glyph origin.
     // part of the glyph to the left of it is displayed in a different color.
-    int be;                     // blur edges
-    double blur;                // gaussian blur
-    double shadow_x;
-    double shadow_y;
-    double frx, fry, frz;       // rotation
-    double fax, fay;            // text shearing
-    double scale_x, scale_y;
-    int border_style;
-    int has_border;
-    double border_x, border_y;
-    double hspacing;
-    unsigned italic;
-    unsigned bold;
-    int flags;
-    int shift_x, shift_y;
 
-    unsigned has_outline;
-    unsigned is_drawing;
-
-    int max_str_length;
-    int str_length;
-    unsigned chars;
-    char *str;
-    int cached;
-    FT_Vector pos_orig;
     int first_pos_x;
+
+    size_t bitmap_count, max_bitmap_count;
+    BitmapRef *bitmaps;
+
+    int x, y;
+    Rectangle rect, rect_o;
+    Bitmap *bm, *bm_o, *bm_s;   // glyphs, outline, shadow bitmaps
+    size_t n_bm, n_bm_o;
 } CombinedBitmapInfo;
 
 // describes a glyph
@@ -166,11 +144,8 @@ typedef struct glyph_info {
 #endif
     double font_size;
     ASS_Drawing *drawing;
-    FT_Outline *outline;
-    FT_Outline *border;
-    Bitmap *bm;                 // glyph bitmap
-    Bitmap *bm_o;               // outline bitmap
-    Bitmap *bm_s;               // shadow bitmap
+    ASS_Outline *outline;
+    ASS_Outline *border;
     FT_BBox bbox;
     FT_Vector pos;
     FT_Vector offset;
@@ -200,10 +175,10 @@ typedef struct glyph_info {
     unsigned bold;
     int flags;
 
-    int bm_run_id;
     int shape_run_id;
 
     BitmapHashKey hash_key;
+    BitmapHashValue *image;
 
     // next glyph in this cluster
     struct glyph_info *next;
@@ -262,14 +237,13 @@ typedef struct {
     int clip_x0, clip_y0, clip_x1, clip_y1;
     char clip_mode;             // 1 = iclip
     char detect_collisions;
-    uint32_t fade;              // alpha from \fad
+    int fade;                   // alpha from \fad
     char be;                    // blur edges
     double blur;                // gaussian blur
     double shadow_x;
     double shadow_y;
     int drawing_scale;          // currently reading: regular text if 0, drawing otherwise
     double pbo;                 // drawing baseline offset
-    ASS_Drawing *drawing;       // current drawing
     ASS_Drawing *clip_drawing;  // clip vector
     int clip_drawing_mode;      // 0 = regular clip, 1 = inverse clip
 
@@ -277,9 +251,6 @@ typedef struct {
     int effect_timing;
     int effect_skip_timing;
 
-    // bitmap run id (used for final bitmap rendering)
-    int bm_run_id;
-
     enum {
         SCROLL_LR,              // left-to-right
         SCROLL_RL,
@@ -298,6 +269,10 @@ typedef struct {
 
     // combination of ASS_OVERRIDE_BIT_* flags that apply right now
     unsigned overrides;
+    // whether to apply font_scale
+    int apply_font_scale;
+    // whether this is assumed to be explicitly positioned
+    int explicit;
 
     // used to store RenderContext.style when doing selective style overrides
     ASS_Style override_style_temp_storage;
@@ -313,24 +288,14 @@ typedef struct {
     size_t composite_max_size;
 } CacheStore;
 
-typedef void (*BitmapBlendFunc)(uint8_t *dst, intptr_t dst_stride,
-                                uint8_t *src, intptr_t src_stride,
-                                intptr_t height, intptr_t width);
-typedef void (*BitmapMulFunc)(uint8_t *dst, intptr_t dst_stride,
-                              uint8_t *src1, intptr_t src1_stride,
-                              uint8_t *src2, intptr_t src2_stride,
-                              intptr_t width, intptr_t height);
-typedef void (*BEBlurFunc)(uint8_t *buf, intptr_t w,
-                           intptr_t h, intptr_t stride,
-                           uint16_t *tmp);
+#include "ass_shaper.h"
 
 struct ass_renderer {
     ASS_Library *library;
     FT_Library ftlibrary;
-    FCInstance *fontconfig_priv;
+    ASS_FontSelector *fontselect;
     ASS_Settings settings;
     int render_id;
-    ASS_SynthPriv *synth_priv;
     ASS_Shaper *shaper;
 
     ASS_Image *images_root;     // rendering result is stored here
@@ -357,13 +322,10 @@ struct ass_renderer {
     TextInfo text_info;
     CacheStore cache;
 
+    const BitmapEngine *engine;
 #if CONFIG_RASTERIZER
-    ASS_Rasterizer rasterizer;
+    RasterizerData rasterizer;
 #endif
-    BitmapBlendFunc add_bitmaps_func;
-    BitmapBlendFunc sub_bitmaps_func;
-    BitmapMulFunc mul_bitmaps_func;
-    BEBlurFunc be_blur_func;
 
     FreeList *free_head;
     FreeList *free_tail;
