@@ -17,18 +17,18 @@
  */
 
 #include "config.h"
+#include "ass_compat.h"
 
 #include <stddef.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <stdint.h>
 #include <inttypes.h>
-#include <strings.h>
-#include <limits.h>
 
 #include "ass_library.h"
 #include "ass.h"
 #include "ass_utils.h"
+#include "ass_string.h"
 
 #if (defined(__i386__) || defined(__x86_64__)) && CONFIG_ASM
 
@@ -38,38 +38,50 @@ int has_sse2(void)
 {
     uint32_t eax = 1, ebx, ecx, edx;
     ass_get_cpuid(&eax, &ebx, &ecx, &edx);
-    return (!!(edx & (1 << 26)));
+    return (edx >> 26) & 0x1;
 }
 
 int has_avx(void)
 {
     uint32_t eax = 1, ebx, ecx, edx;
     ass_get_cpuid(&eax, &ebx, &ecx, &edx);
-    if(!(ecx & (1 << 27))){
+    if(!(ecx & (1 << 27))) // not OSXSAVE
         return 0;
-    }
     uint32_t misc = ecx;
+    ass_get_xgetbv(0, &eax, &edx);
+    if((eax & 0x6) != 0x6)
+        return 0;
     eax = 0;
     ass_get_cpuid(&eax, &ebx, &ecx, &edx);
-    if((ecx & (0x2 | 0x4)) != (0x2 | 0x4)){
-        return 0;
-    }
-    return (!!(misc & (1 << 28)));
+    return (ecx & 0x6) == 0x6 ? (misc >> 28) & 0x1 : 0; // check high bits are relevant, then AVX support
 }
 
 int has_avx2(void)
 {
     uint32_t eax = 7, ebx, ecx, edx;
     ass_get_cpuid(&eax, &ebx, &ecx, &edx);
-    return (!!(ebx & (1 << 5))) && has_avx();
+    return (ebx >> 5) & has_avx();
 }
 
 #endif // ASM
 
+#ifndef HAVE_STRNDUP
+char *ass_strndup(const char *s, size_t n)
+{
+    char *end = memchr(s, 0, n);
+    size_t len = end ? end - s : n;
+    char *new = len < SIZE_MAX ? malloc(len + 1) : NULL;
+    if (new) {
+        memcpy(new, s, len);
+        new[len] = 0;
+    }
+    return new;
+}
+#endif
+
 void *ass_aligned_alloc(size_t alignment, size_t size)
 {
-    if (alignment & (alignment - 1))
-        abort(); // not a power of 2
+    assert(!(alignment & (alignment - 1))); // alignment must be power of 2
     if (size >= SIZE_MAX - alignment - sizeof(void *))
         return NULL;
     char *allocation = malloc(size + sizeof(void *) + alignment - 1);
@@ -89,6 +101,41 @@ void ass_aligned_free(void *ptr)
         free(*((void **)ptr - 1));
 }
 
+/**
+ * This works similar to realloc(ptr, nmemb * size), but checks for overflow.
+ *
+ * Unlike some implementations of realloc, this never acts as a call to free().
+ * If the total size is 0, it is bumped up to 1. This means a NULL return always
+ * means allocation failure, and the unportable realloc(0, 0) case is avoided.
+ */
+void *ass_realloc_array(void *ptr, size_t nmemb, size_t size)
+{
+    if (nmemb > (SIZE_MAX / size))
+        return NULL;
+    size *= nmemb;
+    if (size < 1)
+        size = 1;
+
+    return realloc(ptr, size);
+}
+
+/**
+ * Like ass_realloc_array(), but:
+ * 1. on failure, return the original ptr value, instead of NULL
+ * 2. set errno to indicate failure (errno!=0) or success (errno==0)
+ */
+void *ass_try_realloc_array(void *ptr, size_t nmemb, size_t size)
+{
+    void *new_ptr = ass_realloc_array(ptr, nmemb, size);
+    if (new_ptr) {
+        errno = 0;
+        return new_ptr;
+    } else {
+        errno = ENOMEM;
+        return ptr;
+    }
+}
+
 void skip_spaces(char **str)
 {
     char *p = *str;
@@ -107,92 +154,138 @@ void rskip_spaces(char **str, char *limit)
 
 int mystrtoi(char **p, int *res)
 {
-    double temp_res;
     char *start = *p;
-    temp_res = ass_strtod(*p, p);
+    double temp_res = ass_strtod(*p, p);
     *res = (int) (temp_res + (temp_res > 0 ? 0.5 : -0.5));
-    if (*p != start)
-        return 1;
-    else
-        return 0;
+    return *p != start;
 }
 
 int mystrtoll(char **p, long long *res)
 {
-    double temp_res;
     char *start = *p;
-    temp_res = ass_strtod(*p, p);
+    double temp_res = ass_strtod(*p, p);
     *res = (long long) (temp_res + (temp_res > 0 ? 0.5 : -0.5));
-    if (*p != start)
-        return 1;
-    else
-        return 0;
-}
-
-int mystrtou32(char **p, int base, uint32_t *res)
-{
-    char *start = *p;
-    *res = strtoll(*p, p, base);
-    if (*p != start)
-        return 1;
-    else
-        return 0;
+    return *p != start;
 }
 
 int mystrtod(char **p, double *res)
 {
     char *start = *p;
     *res = ass_strtod(*p, p);
-    if (*p != start)
-        return 1;
-    else
-        return 0;
+    return *p != start;
 }
 
-uint32_t string2color(ASS_Library *library, char *p, int hex)
+int mystrtoi32(char **p, int base, int32_t *res)
+{
+    char *start = *p;
+    long long temp_res = strtoll(*p, p, base);
+    *res = FFMINMAX(temp_res, INT32_MIN, INT32_MAX);
+    return *p != start;
+}
+
+static int read_digits(char **str, int base, uint32_t *res)
+{
+    char *p = *str;
+    char *start = p;
+    uint32_t val = 0;
+
+    while (1) {
+        int digit;
+        if (*p >= '0' && *p < base + '0')
+            digit = *p - '0';
+        else if (*p >= 'a' && *p < base - 10 + 'a')
+            digit = *p - 'a' + 10;
+        else if (*p >= 'A' && *p < base - 10 + 'A')
+            digit = *p - 'A' + 10;
+        else
+            break;
+        val = val * base + digit;
+        ++p;
+    }
+
+    *res = val;
+    *str = p;
+    return p != start;
+}
+
+/**
+ * \brief Convert a string to an integer reduced modulo 2**32
+ * Follows the rules for strtoul but reduces the number modulo 2**32
+ * instead of saturating it to 2**32 - 1.
+ */
+static int mystrtou32_modulo(char **p, int base, uint32_t *res)
+{
+    // This emulates scanf with %d or %x format as it works on
+    // Windows, because that's what is used by VSFilter. In practice,
+    // scanf works the same way on other platforms too, but
+    // the standard leaves its behavior on overflow undefined.
+
+    // Unlike scanf and like strtoul, produce 0 for invalid inputs.
+
+    char *start = *p;
+    int sign = 1;
+
+    skip_spaces(p);
+
+    if (**p == '+')
+        ++*p;
+    else if (**p == '-')
+        sign = -1, ++*p;
+
+    if (base == 16 && !ass_strncasecmp(*p, "0x", 2))
+        *p += 2;
+
+    if (read_digits(p, base, res)) {
+        *res *= sign;
+        return 1;
+    } else {
+        *p = start;
+        return 0;
+    }
+}
+
+int32_t parse_alpha_tag(char *str)
+{
+    int32_t alpha = 0;
+
+    while (*str == '&' || *str == 'H')
+        ++str;
+
+    mystrtoi32(&str, 16, &alpha);
+    return alpha;
+}
+
+uint32_t parse_color_tag(char *str)
+{
+    int32_t color = 0;
+
+    while (*str == '&' || *str == 'H')
+        ++str;
+
+    mystrtoi32(&str, 16, &color);
+    return ass_bswap32((uint32_t) color);
+}
+
+uint32_t parse_color_header(char *str)
 {
     uint32_t color = 0;
-    int base = hex ? 16 : 10;
+    int base;
 
-    if (*p == '&')
-        while (*p == '&')
-            ++p;
-    else
-        ass_msg(library, MSGL_DBG2, "suspicious color format: \"%s\"\n", p);
+    if (!ass_strncasecmp(str, "&h", 2) || !ass_strncasecmp(str, "0x", 2)) {
+        str += 2;
+        base = 16;
+    } else
+        base = 10;
 
-    if (*p == 'H' || *p == 'h') {
-        ++p;
-        mystrtou32(&p, 16, &color);
-    } else {
-        mystrtou32(&p, base, &color);
-    }
-
-    while (*p == '&' || *p == 'H')
-        ++p;
-
-    {
-        unsigned char *tmp = (unsigned char *) (&color);
-        unsigned char b;
-        b = tmp[0];
-        tmp[0] = tmp[3];
-        tmp[3] = b;
-        b = tmp[1];
-        tmp[1] = tmp[2];
-        tmp[2] = b;
-    }
-
-    return color;
+    mystrtou32_modulo(&str, base, &color);
+    return ass_bswap32(color);
 }
 
 // Return a boolean value for a string
 char parse_bool(char *str)
 {
     skip_spaces(&str);
-    if (!strncasecmp(str, "yes", 3))
-        return 1;
-    else if (strtol(str, NULL, 10) > 0)
-        return 1;
-    return 0;
+    return !ass_strncasecmp(str, "yes", 3) || strtol(str, NULL, 10) > 0;
 }
 
 int parse_ycbcr_matrix(char *str)
@@ -209,31 +302,31 @@ int parse_ycbcr_matrix(char *str)
     // so we can simply chop off the rest of the input.
     char buffer[16];
     size_t n = FFMIN(end - str, sizeof buffer - 1);
-    strncpy(buffer, str, n);
+    memcpy(buffer, str, n);
     buffer[n] = '\0';
 
-    if (!strcasecmp(buffer, "none"))
+    if (!ass_strcasecmp(buffer, "none"))
         return YCBCR_NONE;
-    if (!strcasecmp(buffer, "tv.601"))
+    if (!ass_strcasecmp(buffer, "tv.601"))
         return YCBCR_BT601_TV;
-    if (!strcasecmp(buffer, "pc.601"))
+    if (!ass_strcasecmp(buffer, "pc.601"))
         return YCBCR_BT601_PC;
-    if (!strcasecmp(buffer, "tv.709"))
+    if (!ass_strcasecmp(buffer, "tv.709"))
         return YCBCR_BT709_TV;
-    if (!strcasecmp(buffer, "pc.709"))
+    if (!ass_strcasecmp(buffer, "pc.709"))
         return YCBCR_BT709_PC;
-    if (!strcasecmp(buffer, "tv.240m"))
+    if (!ass_strcasecmp(buffer, "tv.240m"))
         return YCBCR_SMPTE240M_TV;
-    if (!strcasecmp(buffer, "pc.240m"))
+    if (!ass_strcasecmp(buffer, "pc.240m"))
         return YCBCR_SMPTE240M_PC;
-    if (!strcasecmp(buffer, "tv.fcc"))
+    if (!ass_strcasecmp(buffer, "tv.fcc"))
         return YCBCR_FCC_TV;
-    if (!strcasecmp(buffer, "pc.fcc"))
+    if (!ass_strcasecmp(buffer, "pc.fcc"))
         return YCBCR_FCC_PC;
     return YCBCR_UNKNOWN;
 }
 
-void ass_msg(ASS_Library *priv, int lvl, char *fmt, ...)
+void ass_msg(ASS_Library *priv, int lvl, const char *fmt, ...)
 {
     va_list va;
     va_start(va, fmt);
@@ -241,6 +334,28 @@ void ass_msg(ASS_Library *priv, int lvl, char *fmt, ...)
     va_end(va);
 }
 
+/**
+ * Return a string with spaces trimmed at start and end.
+ * \param str input string
+ * \return output string, can be released with free()
+ */
+char *strdup_trimmed(const char *str)
+{
+    int left = 0;
+    int right = strlen(str) - 1;
+    char *out = NULL;
+
+    while (ass_isspace(str[left])) left++;
+    while (right > left && ass_isspace(str[right])) right--;
+
+    out = calloc(1, right-left+2);
+
+    if (out)
+        memcpy(out, str + left, right-left+1);
+
+    return out;
+}
+
 unsigned ass_utf8_get_char(char **str)
 {
     uint8_t *strp = (uint8_t *) * str;
@@ -303,6 +418,64 @@ unsigned ass_utf8_put_char(char *dest, uint32_t ch)
     return dest - orig_dest;
 }
 
+/**
+ * \brief Parse UTF-16 and return the code point of the sequence starting at src.
+ * \param src pointer to a pointer to the start of the UTF-16 data
+ *            (will be set to the start of the next code point)
+ * \return the code point
+ */
+static uint32_t ass_read_utf16be(uint8_t **src, size_t bytes)
+{
+    if (bytes < 2)
+        goto too_short;
+
+    uint32_t cp = ((*src)[0] << 8) | (*src)[1];
+    *src += 2;
+    bytes -= 2;
+
+    if (cp >= 0xD800 && cp <= 0xDBFF) {
+        if (bytes < 2)
+            goto too_short;
+
+        uint32_t cp2 = ((*src)[0] << 8) | (*src)[1];
+
+        if (cp2 < 0xDC00 || cp2 > 0xDFFF)
+            return 0xFFFD;
+
+        *src += 2;
+
+        cp = 0x10000 + ((cp - 0xD800) << 10) + (cp2 - 0xDC00);
+    }
+
+    if (cp >= 0xDC00 && cp <= 0xDFFF)
+        return 0xFFFD;
+
+    return cp;
+
+too_short:
+    *src += bytes;
+    return 0xFFFD;
+}
+
+void ass_utf16be_to_utf8(char *dst, size_t dst_size, uint8_t *src, size_t src_size)
+{
+    uint8_t *end = src + src_size;
+
+    if (!dst_size)
+        return;
+
+    while (src < end) {
+        uint32_t cp = ass_read_utf16be(&src, end - src);
+        if (dst_size < 5)
+            break;
+        unsigned s = ass_utf8_put_char(dst, cp);
+        dst += s;
+        dst_size -= s;
+    }
+
+    *dst = '\0';
+}
+
 /**
  * \brief find style by name
  * \param track track
@@ -320,7 +493,7 @@ int lookup_style(ASS_Track *track, char *name)
         ++name;
     // VSFilter then normalizes the case of "Default"
     // (only in contexts where this function is called)
-    if (strcasecmp(name, "Default") == 0)
+    if (ass_strcasecmp(name, "Default") == 0)
         name = "Default";
     for (i = track->n_styles - 1; i >= 0; --i) {
         if (strcmp(track->styles[i].Name, name) == 0)
@@ -355,47 +528,3 @@ ASS_Style *lookup_style_strict(ASS_Track *track, char *name, size_t len)
     return NULL;
 }
 
-#ifdef CONFIG_ENCA
-void *ass_guess_buffer_cp(ASS_Library *library, unsigned char *buffer,
-                          int buflen, char *preferred_language,
-                          char *fallback)
-{
-    const char **languages;
-    size_t langcnt;
-    EncaAnalyser analyser;
-    EncaEncoding encoding;
-    char *detected_sub_cp = NULL;
-    int i;
-
-    languages = enca_get_languages(&langcnt);
-    ass_msg(library, MSGL_V, "ENCA supported languages");
-    for (i = 0; i < langcnt; i++) {
-        ass_msg(library, MSGL_V, "lang %s", languages[i]);
-    }
-
-    for (i = 0; i < langcnt; i++) {
-        const char *tmp;
-
-        if (strcasecmp(languages[i], preferred_language) != 0)
-            continue;
-        analyser = enca_analyser_alloc(languages[i]);
-        encoding = enca_analyse_const(analyser, buffer, buflen);
-        tmp = enca_charset_name(encoding.charset, ENCA_NAME_STYLE_ICONV);
-        if (tmp && encoding.charset != ENCA_CS_UNKNOWN) {
-            detected_sub_cp = strdup(tmp);
-            ass_msg(library, MSGL_INFO, "ENCA detected charset: %s", tmp);
-        }
-        enca_analyser_free(analyser);
-    }
-
-    free(languages);
-
-    if (!detected_sub_cp) {
-        detected_sub_cp = strdup(fallback);
-        ass_msg(library, MSGL_INFO,
-               "ENCA detection failed: fallback to %s", fallback);
-    }
-
-    return detected_sub_cp;
-}
-#endif
